import os
from pathlib import Path
from appdirs import AppDirs

_app_dirs = AppDirs("BinanceWatch", "EtWnn")


def get_data_path():
    """
    Return the folder path where to store the data created by this project
    It uses the library appdirs to follow the conventions across multi OS(MAc, Linux, Windows)

    https://pypi.org/project/appdirs/

    :return: path of the folder to use for data saving
    :rtype: pathlib.Path
    """
    return Path(_app_dirs.user_data_dir)


try:  # create the data folder path
    os.makedirs(get_data_path())
except FileExistsError:
    pass





import datetime
import math
import time
from typing import Optional, Dict

import dateparser
from binance.client import Client
from binance.exceptions import BinanceAPIException
from tqdm import tqdm

from BinanceWatch.storage import tables
from BinanceWatch.utils.LoggerGenerator import LoggerGenerator
from BinanceWatch.utils.time_utils import datetime_to_millistamp
from BinanceWatch.storage.BinanceDataBase import BinanceDataBase


class BinanceManager:
    """
    This class is in charge of filling the database by calling the binance API
    """
    API_MAX_RETRY = 3

    def __init__(self, api_key: str, api_secret: str, account_name: str = 'default'):
        """
        initialise the binance manager.

        :param api_key: key for the Binance api
        :type api_key: str
        :param api_secret: secret for the Binance api
        :type api_secret: str
        :param account_name: if you have several accounts to monitor, you need to give them different names or the
        database will collide
        :type account_name: str
        """
        self.account_name = account_name
        self.db = BinanceDataBase(name=f"{self.account_name}_db")
        self.client = Client(api_key=api_key, api_secret=api_secret)
        self.logger = LoggerGenerator.get_logger(f"BinanceManager_{self.account_name}")

    def update_spot(self):
        """
        call all update methods related to the spot account

        :return: None
        :rtype: None
        """
        self.update_all_spot_trades()
        self.update_spot_deposits()
        self.update_spot_withdraws()
        self.update_spot_dusts()
        self.update_spot_dividends()
        self.update_universal_transfers(transfer_filter='MAIN')

    def update_cross_margin(self):
        """
        call all update methods related to cross margin spot account

        :return: None
        :rtype: None
        """
        self.update_all_cross_margin_trades()
        self.update_cross_margin_loans()
        self.update_cross_margin_interests()
        self.update_cross_margin_repays()
        self.update_universal_transfers(transfer_filter='MARGIN')

    def update_lending(self):
        """
        call all update methods related to lending activities

        :return: None
        :rtype: None
        """
        self.update_lending_interests()
        self.update_lending_purchases()
        self.update_lending_redemptions()

    def update_universal_transfers(self, transfer_filter: Optional[str] = None):
        """
        update the universal transfers database.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.query_universal_transfer_history
        https://binance-docs.github.io/apidocs/spot/en/#query-user-universal-transfer-history

        :param transfer_filter: if not None, only the transfers containing this filter will be updated (ex: 'MAIN')
        :type transfer_filter: Optional[str]
        :return: None
        :rtype: None
        """
        all_types = ['MAIN_C2C', 'MAIN_UMFUTURE', 'MAIN_CMFUTURE', 'MAIN_MARGIN', 'MAIN_MINING', 'C2C_MAIN',
                     'C2C_UMFUTURE', 'C2C_MINING', 'C2C_MARGIN', 'UMFUTURE_MAIN', 'UMFUTURE_C2C',
                     'UMFUTURE_MARGIN', 'CMFUTURE_MAIN', 'CMFUTURE_MARGIN', 'MARGIN_MAIN', 'MARGIN_UMFUTURE',
                     'MARGIN_CMFUTURE', 'MARGIN_MINING', 'MARGIN_C2C', 'MINING_MAIN', 'MINING_UMFUTURE',
                     'MINING_C2C', 'MINING_MARGIN']
        if transfer_filter is not None:
            transfers_types = list(filter(lambda x: transfer_filter in x, all_types))
        else:
            transfers_types = all_types
        pbar = tqdm(total=len(transfers_types))
        for transfer_type in transfers_types:
            pbar.set_description(f"fetching transfer type {transfer_type}")
            latest_time = self.db.get_last_universal_transfer_time(transfer_type=transfer_type) + 1
            current = 1
            while True:
                client_params = {
                    'type': transfer_type,
                    'startTime': latest_time,
                    'current': current,
                    'size': 100
                }
                universal_transfers = self._call_binance_client('query_universal_transfer_history', client_params)

                try:
                    universal_transfers = universal_transfers['rows']
                except KeyError:
                    break
                for transfer in universal_transfers:
                    self.db.add_universal_transfer(transfer_id=transfer['tranId'],
                                                   transfer_type=transfer['type'],
                                                   transfer_time=transfer['timestamp'],
                                                   asset=transfer['asset'],
                                                   amount=float(transfer['amount'])
                                                   )

                if len(universal_transfers):
                    current += 1  # next page
                    self.db.commit()
                else:
                    break
            pbar.update()
        pbar.close()

    def update_cross_margin_interests(self):
        """
        update the interests for all cross margin assets

        sources:
        https://binance-docs.github.io/apidocs/spot/en/#query-repay-record-user_data

        :return:
        :rtype:
        """
        margin_type = 'cross'
        latest_time = self.db.get_last_margin_interest_time(margin_type)
        archived = 1000 * time.time() - latest_time > 1000 * 3600 * 24 * 30 * 3
        current = 1
        pbar = tqdm()
        pbar.set_description("fetching cross margin interests")
        while True:
            params = {
                'current': current,
                'startTime': latest_time + 1000,
                'size': 100,
                'archived': archived
            }

            # no built-in method yet in python-binance for margin/interestHistory
            client_params = {
                'method': 'get',
                'path': 'margin/interestHistory',
                'signed': True,
                'data': params
            }
            interests = self._call_binance_client('_request_margin_api', client_params)

            for interest in interests['rows']:
                self.db.add_margin_interest(margin_type=margin_type,
                                            interest_time=interest['interestAccuredTime'],
                                            asset=interest['asset'],
                                            interest=interest['interest'],
                                            interest_type=interest['type'],
                                            auto_commit=False)

            if len(interests['rows']):
                current += 1  # next page
                self.db.commit()
            elif archived:  # switching to non archived interests
                current = 1
                archived = False
                latest_time = self.db.get_last_margin_interest_time(margin_type)
            else:
                break
            pbar.update()
        pbar.close()

    def update_cross_margin_repays(self):
        """
        update the repays for all cross margin assets

        :return: None
        :rtype: None
        """
        client_params = {
            'method': 'get',
            'path': 'margin/allPairs',
            'data': {}
        }
        symbols_info = self._call_binance_client('_request_margin_api', client_params)  # not built-in yet
        assets = set()
        for symbol_info in symbols_info:
            assets.add(symbol_info['base'])
            assets.add(symbol_info['quote'])

        pbar = tqdm(total=len(assets))
        for asset in assets:
            pbar.set_description(f"fetching {asset} cross margin repays")
            self.update_margin_asset_repay(asset=asset)
            pbar.update()
        pbar.close()

    def update_margin_asset_repay(self, asset: str, isolated_symbol=''):
        """
        update the repays database for a specified asset.

        sources:
        https://binance-docs.github.io/apidocs/spot/en/#query-repay-record-user_data
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_margin_repay_details

        :param asset: asset for the repays
        :type asset: str
        :param isolated_symbol: the symbol must be specified of isolated margin, otherwise cross margin data is returned
        :type isolated_symbol: str
        :return: None
        :rtype: None
        """
        margin_type = 'cross' if isolated_symbol == '' else 'isolated'
        latest_time = self.db.get_last_repay_time(asset=asset, margin_type=margin_type)
        archived = 1000 * time.time() - latest_time > 1000 * 3600 * 24 * 30 * 3
        current = 1
        while True:
            client_params = {
                'asset': asset,
                'current':current,
                'startTime': latest_time + 1000,
                'archived': archived,
                'isolatedSymbol': isolated_symbol,
                'size': 100
            }
            repays = self._call_binance_client('get_margin_repay_details', client_params)

            for repay in repays['rows']:
                if repay['status'] == 'CONFIRMED':
                    self.db.add_repay(margin_type=margin_type,
                                      tx_id=repay['txId'],
                                      repay_time=repay['timestamp'],
                                      asset=repay['asset'],
                                      principal=repay['principal'],
                                      interest=repay['interest'],
                                      auto_commit=False)

            if len(repays['rows']):
                current += 1  # next page
                self.db.commit()
            elif archived:  # switching to non archived repays
                current = 1
                archived = False
                latest_time = self.db.get_last_repay_time(asset=asset, margin_type=margin_type)
            else:
                break

    def update_cross_margin_loans(self):
        """
        update the loans for all cross margin assets

        :return: None
        :rtype: None
        """
        client_params = {
            'method': 'get',
            'path': 'margin/allPairs',
            'data': {}
        }
        symbols_info = self._call_binance_client('_request_margin_api', client_params)  # not built-in yet
        assets = set()
        for symbol_info in symbols_info:
            assets.add(symbol_info['base'])
            assets.add(symbol_info['quote'])

        pbar = tqdm(total=len(assets))
        for asset in assets:
            pbar.set_description(f"fetching {asset} cross margin loans")
            self.update_margin_asset_loans(asset=asset)
            pbar.update()
        pbar.close()

    def update_margin_asset_loans(self, asset: str, isolated_symbol=''):
        """
        update the loans database for a specified asset.

        sources:
        https://binance-docs.github.io/apidocs/spot/en/#query-loan-record-user_data
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_margin_loan_details

        :param asset: asset for the loans
        :type asset: str
        :param isolated_symbol: the symbol must be specified of isolated margin, otherwise cross margin data is returned
        :type isolated_symbol: str
        :return: None
        :rtype: None
        """
        margin_type = 'cross' if isolated_symbol == '' else 'isolated'
        latest_time = self.db.get_last_loan_time(asset=asset, margin_type=margin_type)
        archived = 1000 * time.time() - latest_time > 1000 * 3600 * 24 * 30 * 3
        current = 1
        while True:
            client_params = {
                'asset': asset,
                'current': current,
                'startTime': latest_time + 1000,
                'archived': archived,
                'isolatedSymbol': isolated_symbol,
                'size': 100
            }
            loans = self._call_binance_client('get_margin_loan_details', client_params)

            for loan in loans['rows']:
                if loan['status'] == 'CONFIRMED':
                    self.db.add_loan(margin_type=margin_type,
                                     tx_id=loan['txId'],
                                     loan_time=loan['timestamp'],
                                     asset=loan['asset'],
                                     principal=loan['principal'],
                                     auto_commit=False)

            if len(loans['rows']):
                current += 1  # next page
                self.db.commit()
            elif archived:  # switching to non archived loans
                current = 1
                archived = False
                latest_time = self.db.get_last_loan_time(asset=asset, margin_type=margin_type)
            else:
                break

    def update_cross_margin_symbol_trades(self, asset: str, ref_asset: str, limit: int = 1000):
        """
        This update the cross_margin trades in the database for a single trading pair.
        It will check the last trade id and will requests the all trades after this trade_id.

        sources:
        https://binance-docs.github.io/apidocs/spot/en/#query-margin-account-39-s-trade-list-user_data
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_margin_trades

        :param asset: name of the asset in the trading pair (ex 'BTC' for 'BTCUSDT')
        :type asset: string
        :param ref_asset: name of the reference asset in the trading pair (ex 'USDT' for 'BTCUSDT')
        :type ref_asset: string
        :param limit: max size of each trade requests
        :type limit: int
        :return: None
        :rtype: None
        """
        limit = min(1000, limit)
        symbol = asset + ref_asset
        last_trade_id = self.db.get_max_trade_id(asset, ref_asset, 'cross_margin')
        while True:
            client_params = {
                'symbol': symbol,
                'fromId': last_trade_id + 1,
                'limit': limit
            }
            new_trades = self._call_binance_client('get_margin_trades', client_params)

            for trade in new_trades:
                self.db.add_trade(trade_type='cross_margin',
                                  trade_id=int(trade['id']),
                                  trade_time=int(trade['time']),
                                  asset=asset,
                                  ref_asset=ref_asset,
                                  qty=float(trade['qty']),
                                  price=float(trade['price']),
                                  fee=float(trade['commission']),
                                  fee_asset=trade['commissionAsset'],
                                  is_buyer=trade['isBuyer'],
                                  auto_commit=False
                                  )
                last_trade_id = max(last_trade_id, int(trade['id']))
            if len(new_trades):
                self.db.commit()
            if len(new_trades) < limit:
                break

    def update_all_cross_margin_trades(self, limit: int = 1000):
        """
        This update the cross margin trades in the database for every trading pairs

        :param limit: max size of each trade requests
        :type limit: int
        :return: None
        :rtype: None
        """
        client_params = {
            'method': 'get',
            'path': 'margin/allPairs',
            'data': {}
        }
        symbols_info = self._call_binance_client('_request_margin_api', client_params)  # not built-in yet

        pbar = tqdm(total=len(symbols_info))
        for symbol_info in symbols_info:
            pbar.set_description(f"fetching {symbol_info['symbol']} cross margin trades")
            self.update_cross_margin_symbol_trades(asset=symbol_info['base'],
                                                   ref_asset=symbol_info['quote'],
                                                   limit=limit)
            pbar.update()
        pbar.close()

    def update_lending_redemptions(self):
        """
        update the lending redemptions database.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_lending_redemption_history
        https://binance-docs.github.io/apidocs/spot/en/#get-redemption-record-user_data

        :return: None
        :rtype: None
        """
        lending_types = ['DAILY', 'ACTIVITY', 'CUSTOMIZED_FIXED']
        pbar = tqdm(total=3)
        for lending_type in lending_types:
            pbar.set_description(f"fetching lending redemptions of type {lending_type}")
            latest_time = self.db.get_last_lending_redemption_time(lending_type=lending_type) + 1
            current = 1
            while True:
                client_params = {
                    'lendingType': lending_type,
                    'startTime': latest_time,
                    'current': current,
                    'size': 100
                }
                lending_redemptions = self._call_binance_client('get_lending_redemption_history', client_params)

                for li in lending_redemptions:
                    if li['status'] == 'PAID':
                        self.db.add_lending_redemption(redemption_time=li['createTime'],
                                                       lending_type=lending_type,
                                                       asset=li['asset'],
                                                       amount=li['amount']
                                                       )

                if len(lending_redemptions):
                    current += 1  # next page
                    self.db.commit()
                else:
                    break
            pbar.update()
        pbar.close()

    def update_lending_purchases(self):
        """
        update the lending purchases database.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_lending_purchase_history
        https://binance-docs.github.io/apidocs/spot/en/#get-purchase-record-user_data

        :return: None
        :rtype: None
        """
        lending_types = ['DAILY', 'ACTIVITY', 'CUSTOMIZED_FIXED']
        pbar = tqdm(total=3)
        for lending_type in lending_types:
            pbar.set_description(f"fetching lending purchases of type {lending_type}")
            latest_time = self.db.get_last_lending_purchase_time(lending_type=lending_type) + 1
            current = 1
            while True:
                client_params = {
                    'lendingType': lending_type,
                    'startTime': latest_time,
                    'current': current,
                    'size': 100
                }
                lending_purchases = self._call_binance_client('get_lending_purchase_history', client_params)

                for li in lending_purchases:
                    if li['status'] == 'SUCCESS':
                        self.db.add_lending_purchase(purchase_id=li['purchaseId'],
                                                     purchase_time=li['createTime'],
                                                     lending_type=li['lendingType'],
                                                     asset=li['asset'],
                                                     amount=li['amount']
                                                     )

                if len(lending_purchases):
                    current += 1  # next page
                    self.db.commit()
                else:
                    break
            pbar.update()
        pbar.close()

    def update_lending_interests(self):
        """
        update the lending interests database.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_lending_interest_history
        https://binance-docs.github.io/apidocs/spot/en/#get-interest-history-user_data-2

        :return: None
        :rtype: None
        """
        lending_types = ['DAILY', 'ACTIVITY', 'CUSTOMIZED_FIXED']
        pbar = tqdm(total=3)
        for lending_type in lending_types:
            pbar.set_description(f"fetching lending interests of type {lending_type}")
            latest_time = self.db.get_last_lending_interest_time(lending_type=lending_type) + 3600 * 1000  # add 1 hour
            current = 1
            while True:
                client_params = {
                    'lendingType': lending_type,
                    'startTime': latest_time,
                    'current': current,
                    'size': 100
                }
                lending_interests = self._call_binance_client('get_lending_interest_history', client_params)

                for li in lending_interests:
                    self.db.add_lending_interest(time=li['time'],
                                                 lending_type=li['lendingType'],
                                                 asset=li['asset'],
                                                 amount=li['interest']
                                                 )

                if len(lending_interests):
                    current += 1  # next page
                    self.db.commit()
                else:
                    break
            pbar.update()
        pbar.close()

    def update_spot_dusts(self):
        """
        update the dust database. As there is no way to get the dust by id or timeframe, the table is cleared
        for each update

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_dust_log
        https://binance-docs.github.io/apidocs/spot/en/#dustlog-user_data

        :return: None
        :rtype: None
        """
        self.db.drop_table(tables.SPOT_DUST_TABLE)

        result = self._call_binance_client('get_dust_log')
        dusts = result['results']
        pbar = tqdm(total=dusts['total'])
        pbar.set_description("fetching spot dusts")
        for d in dusts['rows']:
            for sub_dust in d['logs']:
                date_time = dateparser.parse(sub_dust['operateTime'] + 'Z')
                self.db.add_dust(tran_id=sub_dust['tranId'],
                                 time=datetime_to_millistamp(date_time),
                                 asset=sub_dust['fromAsset'],
                                 asset_amount=sub_dust['amount'],
                                 bnb_amount=sub_dust['transferedAmount'],
                                 bnb_fee=sub_dust['serviceChargeAmount'],
                                 auto_commit=False
                                 )
            pbar.update()
        self.db.commit()
        pbar.close()

    def update_spot_dividends(self, day_jump: float = 90, limit: int = 500):
        """
        update the dividends database (earnings distributed by Binance)
        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_asset_dividend_history
        https://binance-docs.github.io/apidocs/spot/en/#asset-dividend-record-user_data

        :param day_jump: length of the time window in days, max is 90
        :type day_jump: float
        :param limit: max number of dividends to retrieve per call, max is 500
        :type limit: int
        :return: None
        :rtype: None
        """
        limit = min(500, limit)
        delta_jump = min(day_jump, 90) * 24 * 3600 * 1000
        start_time = self.db.get_last_spot_dividend_time() + 1
        now_millistamp = datetime_to_millistamp(datetime.datetime.now(tz=datetime.timezone.utc))
        pbar = tqdm(total=math.ceil((now_millistamp - start_time) / delta_jump))
        pbar.set_description("fetching spot dividends")
        while start_time < now_millistamp:
            # the stable working version of client.get_asset_dividend_history is not released yet,
            # for now it has a post error, so this protected member is used in the meantime
            params = {
                'startTime': start_time,
                'endTime': start_time + delta_jump,
                'limit': limit
            }
            client_params = {
                'method': 'get',
                'path': 'asset/assetDividend',
                'signed': True,
                'data': params
            }
            result = self._call_binance_client('_request_margin_api', client_params)

            dividends = result['rows']
            for div in dividends:
                self.db.add_dividend(div_id=int(div['tranId']),
                                     div_time=int(div['divTime']),
                                     asset=div['asset'],
                                     amount=float(div['amount']),
                                     auto_commit=False
                                     )
            pbar.update()
            if len(dividends) < limit:
                start_time += delta_jump
            else:  # limit was reached before the end of the time windows
                start_time = int(dividends[0]['divTime']) + 1
            if len(dividends):
                self.db.commit()
        pbar.close()

    def update_spot_withdraws(self, day_jump: float = 90):
        """
        This fetch the crypto withdraws made on the spot account from the last withdraw time in the database to now.
        It is done with multiple call, each having a time window of day_jump days.
        The withdraws are then saved in the database.
        Only successful withdraws are fetched.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_withdraw_history
        https://binance-docs.github.io/apidocs/spot/en/#withdraw-history-user_data

        :param day_jump: length of the time window for each call (max 90)
        :type day_jump: float
        :return: None
        :rtype: None
        """
        delta_jump = min(day_jump, 90) * 24 * 3600 * 1000
        start_time = self.db.get_last_spot_withdraw_time() + 1
        now_millistamp = datetime_to_millistamp(datetime.datetime.now(tz=datetime.timezone.utc))
        pbar = tqdm(total=math.ceil((now_millistamp - start_time) / delta_jump))
        pbar.set_description("fetching spot withdraws")
        while start_time < now_millistamp:
            client_params = {
                'startTime': start_time,
                'endTime': start_time + delta_jump,
                'status': 6
            }
            result = self._call_binance_client('get_withdraw_history', client_params)

            withdraws = result['withdrawList']
            for withdraw in withdraws:
                self.db.add_withdraw(withdraw_id=withdraw['id'],
                                     tx_id=withdraw['txId'],
                                     apply_time=int(withdraw['applyTime']),
                                     asset=withdraw['asset'],
                                     amount=float(withdraw['amount']),
                                     fee=float(withdraw['transactionFee']),
                                     auto_commit=False
                                     )
            pbar.update()
            start_time += delta_jump
            if len(withdraws):
                self.db.commit()
        pbar.close()

    def update_spot_deposits(self, day_jump: float = 90):
        """
        This fetch the crypto deposit made on the spot account from the last deposit time in the database to now.
        It is done with multiple call, each having a time window of day_jump days.
        The deposits are then saved in the database.
        Only successful deposits are fetched.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_deposit_history
        https://binance-docs.github.io/apidocs/spot/en/#deposit-history-user_data

        :param day_jump: length of the time window for each call (max 90)
        :type day_jump: float
        :return: None
        :rtype: None
        """
        delta_jump = min(day_jump, 90) * 24 * 3600 * 1000
        start_time = self.db.get_last_spot_deposit_time() + 1
        now_millistamp = datetime_to_millistamp(datetime.datetime.now(tz=datetime.timezone.utc))
        pbar = tqdm(total=math.ceil((now_millistamp - start_time) / delta_jump))
        pbar.set_description("fetching spot deposits")
        while start_time < now_millistamp:
            client_params = {
                'startTime': start_time,
                'endTime': start_time + delta_jump,
                'status': 1
            }
            result = self._call_binance_client('get_deposit_history', client_params)

            deposits = result['depositList']
            for deposit in deposits:
                self.db.add_deposit(tx_id=deposit['txId'],
                                    asset=deposit['asset'],
                                    insert_time=int(deposit['insertTime']),
                                    amount=float(deposit['amount']),
                                    auto_commit=False)
            pbar.update()
            start_time += delta_jump
            if len(deposits):
                self.db.commit()
        pbar.close()

    def update_spot_symbol_trades(self, asset: str, ref_asset: str, limit: int = 1000):
        """
        This update the spot trades in the database for a single trading pair. It will check the last trade id and will
        requests the all trades after this trade_id.

        sources:
        https://python-binance.readthedocs.io/en/latest/binance.html#binance.client.Client.get_my_trades
        https://binance-docs.github.io/apidocs/spot/en/#account-trade-list-user_data

        :param asset: name of the asset in the trading pair (ex 'BTC' for 'BTCUSDT')
        :type asset: string
        :param ref_asset: name of the reference asset in the trading pair (ex 'USDT' for 'BTCUSDT')
        :type ref_asset: string
        :param limit: max size of each trade requests
        :type limit: int
        :return: None
        :rtype: None
        """
        limit = min(1000, limit)
        symbol = asset + ref_asset
        last_trade_id = self.db.get_max_trade_id(asset, ref_asset, 'spot')
        while True:
            client_params = {
                'symbol': symbol,
                'fromId': last_trade_id + 1,
                'limit': limit
            }
            new_trades = self._call_binance_client('get_my_trades', client_params)

            for trade in new_trades:
                self.db.add_trade(trade_type='spot',
                                  trade_id=int(trade['id']),
                                  trade_time=int(trade['time']),
                                  asset=asset,
                                  ref_asset=ref_asset,
                                  qty=float(trade['qty']),
                                  price=float(trade['price']),
                                  fee=float(trade['commission']),
                                  fee_asset=trade['commissionAsset'],
                                  is_buyer=trade['isBuyer'],
                                  auto_commit=False
                                  )
                last_trade_id = max(last_trade_id, int(trade['id']))
            if len(new_trades):
                self.db.commit()
            if len(new_trades) < limit:
                break

    def update_all_spot_trades(self, limit: int = 1000):
        """
        This update the spot trades in the database for every trading pairs

        :param limit: max size of each trade requests
        :type limit: int
        :return: None
        :rtype: None
        """
        symbols_info = self.client.get_exchange_info()['symbols']
        pbar = tqdm(total=len(symbols_info))
        for symbol_info in symbols_info:
            pbar.set_description(f"fetching {symbol_info['symbol']} spot trades")
            self.update_spot_symbol_trades(asset=symbol_info['baseAsset'],
                                           ref_asset=symbol_info['quoteAsset'],
                                           limit=limit)
            pbar.update()
        pbar.close()

    def _call_binance_client(self, method_name: str, params: Optional[Dict] = None, retry_count: int = 0):
        """
        This method is used to handle rate limits: if a rate limits is breached, it will wait the necessary time
        to call again the API.

        :param method_name: name of the method binance.Client to call
        :type method_name: str
        :param params: parameters to pass to the above method
        :type params: Dict
        :param retry_count: internal use only to count the number of retry if rate limits are breached
        :type retry_count: int
        :return: response of binance.Client method
        :rtype: Dict
        """
        if params is None:
            params = dict()
        if retry_count >= BinanceManager.API_MAX_RETRY:
            raise RuntimeError(f"The API rate limits has been breached {retry_count} times")

        try:
            return getattr(self.client, method_name)(**params)
        except BinanceAPIException as err:
            if err.code == -1003:  # API rate Limits
                wait_time = float(err.response.headers['Retry-After'])
                if err.response.status_code == 418:  # ban
                    self.logger.error(f"API calls resulted in a ban, retry in {wait_time} seconds")
                    raise err
                self.logger.info(f"API calls resulted in a breach of rate limits, will retry after {wait_time} seconds")
                time.sleep(wait_time + 1)
                return self._call_binance_client(method_name, params, retry_count + 1)
            raise err



from BinanceWatch.storage.DataBase import DataBase, SQLConditionEnum
from BinanceWatch.storage.tables import Table

db = DataBase("test_table")

table1 = Table(
    'test_table1',
    [
        'age',
        'surname',
        'weight'
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL'
    ],
    primary_key='key',
    primary_key_sql_type='INTEGER'
)

table2 = Table(
    'test_table2',
    [
        'age',
        'surname',
        'weight'
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL'
    ],
)


def test_create_cmd(verbose=0, **kwargs):
    create_cmd = db.get_create_cmd(table1)
    if verbose:
        print(create_cmd)
    assert create_cmd == "CREATE TABLE test_table1\n([key] INTEGER PRIMARY KEY, [age] INTEGER, [surname] TEXT," \
                         " [weight] REAL)"

    create_cmd = db.get_create_cmd(table2)
    if verbose:
        print(create_cmd)
    assert create_cmd == "CREATE TABLE test_table2\n([age] INTEGER, [surname] TEXT, [weight] REAL)"


def test_inserts_search_1(verbose=0, **kwargs):
    db.drop_table(table1)
    rows = [
        (1, 15, 'Karl', 55.5),
        (2, 18, 'Kitty', 61.1),
        (3, 18, 'Marc', 48.1),
        (8, 55, 'Jean', 78.1)
    ]
    db.add_rows(table1, rows)

    retrieved_row = db.get_row_by_key(table1, 1)
    if verbose:
        print(f"retrieved the row: {retrieved_row} when looking for the row: {rows[0]}")
    assert rows[0] == retrieved_row

    conditions = [
        (table1.age, SQLConditionEnum.equal, 18),
        (table1.weight, SQLConditionEnum.greater_equal, 55),
    ]
    retrieved_rows = db.get_conditions_rows(table1, conditions_list=conditions)
    assert rows[1:2] == retrieved_rows

    # check max weight is right
    assert max([r[-1] for r in rows]) == db.get_conditions_rows(table1, selection=f"MAX({table1.weight})")[0][0]


def test_inserts_search_2(verbose=0, **kwargs):
    db.drop_table(table2)
    rows = [
        (15, 'Karl', 55.5),
        (18, 'Kitty', 61.1),
        (18, 'Kitty', 61.1),  # no set primary key -> should allow identical rows
        (18, 'Marcus', 58.2),
        (18, 'Marc', 48.1),
        (55, 'Jean', 78.1)
    ]
    db.add_rows(table2, rows)

    try:
        db.get_row_by_key(table2, 1)
        raise RuntimeError("the above line should throw an error as no primary key is defined")
    except ValueError:
        pass

    conditions = [
        (table2.age, SQLConditionEnum.equal, 18),
        (table2.weight, SQLConditionEnum.greater_equal, 55),
    ]
    order_list = [table2.weight]
    retrieved_rows = db.get_conditions_rows(table2, conditions_list=conditions, order_list=order_list)
    sub_rows = list(rows[1:4])
    if verbose:
        print("rows retrieved:")
        for row in retrieved_rows:
            print('\t', row)
    sub_rows.sort(key=lambda x: x[-1])
    assert sub_rows == retrieved_rows

    # check min ge is right
    assert max([r[0] for r in rows]) == db.get_conditions_rows(table2, selection=f"MAX({table2.age})")[0][0]



from enum import Enum
from typing import List, Tuple, Optional, Any, Union
import sqlite3

from BinanceWatch.storage.tables import Table
from BinanceWatch.utils.LoggerGenerator import LoggerGenerator
from BinanceWatch.utils.paths import get_data_path


class SQLConditionEnum(Enum):
    """
    https://www.techonthenet.com/sqlite/comparison_operators.php
    """
    equal = '='
    greater_equal = '>='
    greater = '>'
    lower = '<'
    lower_equal = '<='
    diff = '!='


class DataBase:
    """
    This class will be used to interact with sqlite3 databases without having to generates sqlite commands
    """

    def __init__(self, name: str):
        self.name = name
        self.logger = LoggerGenerator.get_logger(self.name)
        self.save_path = get_data_path() / f"{name}.db"
        self.db_conn = sqlite3.connect(self.save_path)
        self.db_cursor = self.db_conn.cursor()

    def _fetch_rows(self, execution_cmd: str):
        """
        execute a command to fetch some rows and return them
        :param execution_cmd: the command to execute
        :return:
        """
        rows = []
        try:
            self.db_cursor.execute(execution_cmd)
        except sqlite3.OperationalError:
            return rows
        while True:
            row = self.db_cursor.fetchone()
            if row is None:
                break
            rows.append(row)
        return rows

    def get_row_by_key(self, table: Table, key_value) -> Optional[Tuple]:
        """
        get the row identified by a primary key value from a table
        :param table: table to fetch the row from
        :param key_value: key value of the row
        :return: None or the row of value
        """
        if table.primary_key is None:
            raise ValueError(f"table {table.name} has no explicit primary key")
        conditions_list = [(table.primary_key, SQLConditionEnum.equal, key_value)]
        rows = self.get_conditions_rows(table, conditions_list=conditions_list)
        if len(rows):
            return rows[0]

    def get_conditions_rows(self, table: Table,
                            selection: Optional[Union[str, List[str]]] = None,
                            conditions_list: Optional[List[Tuple[str, SQLConditionEnum, Any]]] = None,
                            order_list: Optional[List[str]] = None) -> List:
        if selection is None:
            selection = '*'
        elif isinstance(selection, List):
            selection = ','.join(selection)
        if conditions_list is None:
            conditions_list = []
        if order_list is None:
            order_list = []
        execution_cmd = f"SELECT {selection} from {table.name}"
        execution_cmd = self._add_conditions(execution_cmd, conditions_list=conditions_list)
        execution_cmd = self._add_order(execution_cmd, order_list=order_list)
        return self._fetch_rows(execution_cmd)

    def get_all_rows(self, table: Table) -> List:
        return self.get_conditions_rows(table)

    def add_row(self, table: Table, row: Tuple, auto_commit: bool = True, update_if_exists: bool = False):
        row_s = ", ".join(f"'{v}'" for v in row)
        row_s = f'({row_s})'
        execution_order = f"INSERT INTO {table.name} VALUES {row_s}"
        try:
            self.db_cursor.execute(execution_order)
            if auto_commit:
                self.commit()
        except sqlite3.OperationalError:
            self.create_table(table)
            self.db_cursor.execute(execution_order)
            if auto_commit:
                self.commit()
        except sqlite3.IntegrityError as err:
            if update_if_exists:
                self.update_row(table, row, auto_commit)
            else:
                raise err

    def add_rows(self, table: Table, rows: List[Tuple], auto_commit: bool = True, update_if_exists: bool = False):
        for row in rows:
            self.add_row(table, row, auto_commit=False, update_if_exists=update_if_exists)
        if auto_commit:
            self.commit()

    def update_row(self, table: Table, row: Tuple, auto_commit=True):
        row_s = ", ".join(f"{n} = {v}" for n, v in zip(table.columns_names, row))
        execution_order = f"UPDATE {table.name} SET {row_s} WHERE {table.primary_key} = {row[0]}"
        self.db_cursor.execute(execution_order)
        if auto_commit:
            self.commit()

    def create_table(self, table: Table):
        """
        create a table in the database
        :param table: Table instance with the config of the table to create
        :return:
        """
        create_cmd = self.get_create_cmd(table)
        self.db_cursor.execute(create_cmd)
        self.db_conn.commit()

    def drop_table(self, table: Union[Table, str]):
        """
         delete a table from the database

        :param table: table or table name to drop
        :type table: str or Table instance
        :return: None
        :rtype: None
        """
        if isinstance(table, Table):
            table = table.name
        execution_order = f"DROP TABLE IF EXISTS {table}"
        self.db_cursor.execute(execution_order)
        self.db_conn.commit()

    def drop_all_tables(self):
        """
        drop all the tables existing in the database

        :return: None
        :rtype: None
        """
        tables_desc = self.get_all_tables()
        for table_desc in tables_desc:
            self.drop_table(table_desc[1])
        self.commit()

    def get_all_tables(self) -> List[Tuple]:
        """
        return all the tables existing in the database

        :return: tables descriptions
        :rtype: List[Tuple]
        """
        cmd = "SELECT * FROM sqlite_master WHERE type='table';"
        return self._fetch_rows(cmd)

    def commit(self):
        """
        submit and save the database state
        :return:
        """
        self.db_conn.commit()

    @staticmethod
    def _add_conditions(execution_cmd: str, conditions_list: List[Tuple[str, SQLConditionEnum, Any]]):
        """
        add a list of condition to an SQL command
        :param execution_cmd: SQL command without 'WHERE' statement
        :type execution_cmd: str
        :param conditions_list:
        :return:
        """
        if len(conditions_list):
            add_cmd = ' WHERE'
            for column_name, condition, value in conditions_list:
                add_cmd = add_cmd + f" {column_name} {condition.value} '{value}' AND"
            return execution_cmd + add_cmd[:-4]
        else:
            return execution_cmd

    @staticmethod
    def _add_order(execution_cmd: str, order_list: List[str]):
        """
        add an order specification to an SQL command

        :param execution_cmd: SQL command without 'ORDER BY' statement
        :type execution_cmd: str
        :param order_list:
        :type order_list:
        :return:
        :rtype:
        """
        if len(order_list):
            add_cmd = ' ORDER BY'
            for column_name in order_list:
                add_cmd = add_cmd + f" {column_name},"
            return execution_cmd + add_cmd[:-1] + ' ASC'
        else:
            return execution_cmd

    @staticmethod
    def get_create_cmd(table: Table):
        """
        return the command in string format to create a table in the database
        :param table: Table instance with the config if the table to create
        :return: execution command for the table creation
        """
        cmd = ""
        if table.primary_key is not None:
            cmd = f"[{table.primary_key}] {table.primary_key_sql_type} PRIMARY KEY, "
        for arg_name, arg_type in zip(table.columns_names, table.columns_sql_types):
            cmd = cmd + f"[{arg_name}] {arg_type}, "
        return f"CREATE TABLE {table.name}\n({cmd[:-2]})"



import os
import platform
from enum import Enum
from collections.abc import Iterable

if platform.system() == "Windows":
    os.system('color')  # enable colors in consoles


class ColorEnum(Enum):
    """
    Class to store different values to get colors in a console print
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def color_string(s: str, color: ColorEnum):
    return color.value + s + ColorEnum.ENDC.value


def s_to_warning(message: str):
    return color_string(f"WARNING: {message}", ColorEnum.WARNING)


def s_to_fail(message: str):
    return color_string(message, ColorEnum.FAIL)


def s_to_pass(message: str):
    return color_string(message, ColorEnum.OKGREEN)


def get_blue_sep(n: int = 70):
    return color_string(n * '-', ColorEnum.OKBLUE)


def get_blue_title(tester_name: str, method_name: str, n: int = 10):
    message = f"{n * '#'} {tester_name}.{method_name} {n * '#'}"
    return color_string(message, ColorEnum.OKBLUE)


def string_status(instance):
    """
    transform all the attributes and the values of the instance of a class in a string
    :param instance:
    :return: None
    """
    if isinstance(instance, Iterable) and not isinstance(instance, str):
        s = "[\n" + "\n".join([string_status(e) for e in instance]) + "\n]"
    else:
        if '__class__' not in dir(instance):
            s = str(instance)
        else:
            attribute_names = [m for m in dir(instance) if not m.startswith('__') and not callable(getattr(instance, m))]
            if len(attribute_names):
                s = ""
                for attribute_name in attribute_names:
                    attribute = getattr(instance, attribute_name)
                    if isinstance(attribute, Enum):
                        attribute = attribute.name
                    s = s + f"{attribute_name} -> {attribute}\n"
            else:
                s = str(instance)
    return s


def print_status(instance):
    """
    print all the attributes and the values of the instance of a class
    :param instance:
    :return: None
    """
    print(string_status(instance))



import logging
import os
from typing import Optional

from BinanceWatch.utils.paths import get_data_path


class LoggerGenerator:
    """
    This class is a utility to facilitate the creation of loggers for the different classes / files
    """
    LOGS_FOLDER_PATH = get_data_path() / "logs"

    _default_log_level = logging.WARNING
    _default_write_file = False
    _logger_count = 0

    @staticmethod
    def set_global_log_level(log_level: int):
        """
        set the default log level for loggers creation

        :param log_level: threshold to display the message
        :type log_level: logging enum (ex: logging.WARNING)
        :return: None
        :rtype: None
        """
        LoggerGenerator._default_log_level = log_level

    @staticmethod
    def set_default_write_file(write_file: bool):
        """
        set the default write level for loggers creation

        :param write_file: if the logger should save the message in a file
        :type write_file: bool
        :return: None
        :rtype: None
        """
        LoggerGenerator._default_write_file = write_file

    @staticmethod
    def get_logger(logger_name: str, write_file: Optional[bool] = None,
                   log_level: Optional[int] = None) -> logging.Logger:
        """
        create a logger that will display messages according to the log level threshold. If specified, it will
        also save the messages in a file inside the logs folder

        :param logger_name: name of the logger (a unique logger id will be added after the name)
        :type logger_name: str
        :param write_file: if the logger should save the message in a file
        :type write_file: bool
        :param log_level: threshold to display the message
        :type log_level: logging enum (ex: logging.WARNING)
        :return: the logger object
        :rtype: logging.Logger
        """
        if log_level is None:
            log_level = LoggerGenerator._default_log_level
        if write_file is None:
            write_file = LoggerGenerator._default_write_file

        # create logger
        logger = logging.getLogger(f"lg_{LoggerGenerator._logger_count}_{logger_name}")
        logger.setLevel(level=log_level)
        LoggerGenerator._logger_count += 1

        # create formatter and add it to the handlers
        log_format = '[%(asctime)s %(name)s %(levelname)s] %(message)s [%(pathname)s:%(lineno)d in %(funcName)s]'
        formatter = logging.Formatter(log_format)

        if write_file:
            # create file handler for logger.
            log_file_path = LoggerGenerator.LOGS_FOLDER_PATH / f"{logger_name}.log"
            fh = logging.FileHandler(log_file_path)
            fh.setLevel(level=log_level)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

        # create console handler for logger.
        ch = logging.StreamHandler()
        ch.setLevel(level=log_level)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        return logger


try:  # create logs folder
    os.makedirs(LoggerGenerator.LOGS_FOLDER_PATH)
except FileExistsError:
    pass



import datetime
from typing import Optional

from BinanceWatch.storage.DataBase import DataBase, SQLConditionEnum
from BinanceWatch.storage import tables
from BinanceWatch.utils.time_utils import datetime_to_millistamp


class BinanceDataBase(DataBase):
    """
    Handles the recording of the binance account in a local database
    """

    def __init__(self, name: str = 'binance_db'):
        super().__init__(name)

    def add_universal_transfer(self, transfer_id: int, transfer_type: str, transfer_time: int, asset: str,
                               amount: float, auto_commit: bool = True):
        """
        add a universal transfer to the database

        :param transfer_id: id of the transfer
        :type transfer_id:  int
        :param transfer_type: enum of the transfer type (ex: 'MAIN_MARGIN')
        :type transfer_type: str
        :param transfer_time: millistamp of the operation
        :type transfer_time: int
        :param asset: asset that got transferred
        :type asset: str
        :param amount: amount transferred
        :type amount: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        table = tables.UNIVERSAL_TRANSFER_TABLE

        row = (transfer_id, transfer_type, transfer_time, asset, amount)
        self.add_row(table, row, auto_commit=auto_commit)

    def get_universal_transfers(self, transfer_type: Optional[str] = None, asset: Optional[str] = None,
                                start_time: Optional[int] = None, end_time: Optional[int] = None):
        """
        return universal transfers stored in the database. Transfer type, Asset type and time filters can be used

        :param transfer_type: enum of the transfer type (ex: 'MAIN_MARGIN')
        :type transfer_type: Optional[str]
        :param asset: fetch only interests in this asset
        :type asset: Optional[str]
        :param start_time: fetch only interests after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only interests before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (1206491332,        # transfer id
                'MAIN_MARGIN',      # transfer type
                1589121841000,      # time
                'BNB',              # asset
                10.594112),         # amount
            ]
        """
        table = tables.UNIVERSAL_TRANSFER_TABLE

        conditions_list = []
        if transfer_type is not None:
            conditions_list.append((table.trfType,
                                    SQLConditionEnum.equal,
                                    transfer_type))
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.trfTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.trfTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_universal_transfer_time(self, transfer_type: str) -> int:
        """
        return the latest time when a universal transfer was made
        If None, return the millistamp corresponding to 2017/01/01

        :param transfer_type: enum of the transfer type (ex: 'MAIN_MARGIN')
        :type transfer_type: str
        :return: millistamp
        :rtype: int
        """
        table = tables.UNIVERSAL_TRANSFER_TABLE
        conditions_list = [(table.trfType,
                            SQLConditionEnum.equal,
                            transfer_type)]
        selection = f"MAX({table.trfTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_margin_interest(self, margin_type: str, interest_time: int, asset: str, interest: float,
                            interest_type: str, auto_commit: bool = True):
        """
        add a repay to the database

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type: str
        :param interest_time: millistamp of the operation
        :type interest_time: int
        :param asset: asset that got repaid
        :type asset: str
        :param interest: amount of interest accrued
        :type interest: float
        :param interest_type: one of (PERIODIC, ON_BORROW, PERIODIC_CONVERTED, ON_BORROW_CONVERTED)
        :type interest_type: str
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_INTEREST_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        row = (interest_time, asset, interest, interest_type)
        self.add_row(table, row, auto_commit=auto_commit)

    def get_margin_interests(self, margin_type: str, asset: Optional[str] = None, start_time: Optional[int] = None,
                             end_time: Optional[int] = None):
        """
        return margin interests stored in the database. Asset type and time filters can be used

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :param asset: fetch only interests in this asset
        :type asset: Optional[str]
        :param start_time: fetch only interests after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only interests before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                1559415215400,             # time
                'BNB',                     # asset
                0.51561,                   # interest
                'PERIODIC_CONVERTED'),     # interest type
            ]
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_INTEREST_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = []
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.interestTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.interestTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_margin_interest_time(self, margin_type: str, asset: Optional[str] = None):
        """
        return the latest time when a margin interest was accured on a defined asset or on all assets
        If None, return the millistamp corresponding to 2017/01/01

        :param asset: name of the asset charged as interest
        :type asset: Optional[str]
        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :return: millistamp
        :rtype: int
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_INTEREST_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = []
        if asset is not None:
            conditions_list = [(table.asset,
                                SQLConditionEnum.equal,
                                asset)]
        selection = f"MAX({table.interestTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_repay(self, margin_type: str, tx_id: int, repay_time: int, asset: str, principal: float,
                  interest: float, auto_commit: bool = True):
        """
        add a repay to the database

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :param tx_id: binance id for the transaction (uniqueness?)
        :type tx_id: int
        :param repay_time: millitstamp of the operation
        :type repay_time: int
        :param asset: asset that got repaid
        :type asset: str
        :param principal: principal amount repaid for the loan
        :type principal: float
        :param interest: amount of interest repaid for the loan
        :type interest:
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_REPAY_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        row = (tx_id, repay_time, asset, principal, interest)
        self.add_row(table, row, auto_commit=auto_commit)

    def get_repays(self, margin_type: str, asset: Optional[str] = None, start_time: Optional[int] = None,
                   end_time: Optional[int] = None):
        """
        return repays stored in the database. Asset type and time filters can be used

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :param asset: fetch only repays of this asset
        :type asset: Optional[str]
        :param start_time: fetch only repays after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only repays before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (8289451654,        # transaction id
                1559415215400,      # time
                'USDT',             # asset
                145.5491462,        # principal
                0.51561),           # interest
            ]
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_REPAY_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = []
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.repayTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.repayTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_repay_time(self, asset: str, margin_type: str) -> int:
        """
        return the latest time when a repay was made on a defined asset
        If None, return the millistamp corresponding to 2017/01/01

        :param asset: name of the asset repaid
        :type asset: str
        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :return: millistamp
        :rtype: int
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_REPAY_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = [(table.asset,
                            SQLConditionEnum.equal,
                            asset)]
        selection = f"MAX({table.repayTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_loan(self, margin_type: str, tx_id: int, loan_time: int, asset: str, principal: float,
                 auto_commit: bool = True):
        """
        add a loan to the database

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :param tx_id: binance id for the transaction (uniqueness?)
        :type tx_id: int
        :param loan_time: millitstamp of the operation
        :type loan_time: int
        :param asset: asset that got loaned
        :type asset: str
        :param principal: amount of loaned asset
        :type principal: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_LOAN_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        row = (tx_id, loan_time, asset, principal)
        self.add_row(table, row, auto_commit=auto_commit)

    def get_loans(self, margin_type: str, asset: Optional[str] = None, start_time: Optional[int] = None,
                  end_time: Optional[int] = None):
        """
        return loans stored in the database. Asset type and time filters can be used

        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :param asset: fetch only loans of this asset
        :type asset: Optional[str]
        :param start_time: fetch only loans after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only loans before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (8289451654,        # transaction id
                1559415215400,      # time
                'USDT',             # asset
                145.5491462),       # amount
            ]
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_LOAN_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = []
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.loanTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.loanTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_loan_time(self, asset: str, margin_type: str) -> int:
        """
        return the latest time when an loan was made on a defined asset
        If None, return the millistamp corresponding to 2017/01/01

        :param asset: name of the asset loaned
        :type asset: str
        :param margin_type: either 'cross' or 'isolated'
        :type margin_type:
        :return: millistamp
        :rtype: int
        """
        if margin_type == 'cross':
            table = tables.CROSS_MARGIN_LOAN_TABLE
        elif margin_type == 'isolated':
            raise NotImplementedError
        else:
            raise ValueError(f"margin type should be 'cross' or 'isolated' but {margin_type} was received")

        conditions_list = [(table.asset,
                            SQLConditionEnum.equal,
                            asset)]
        selection = f"MAX({table.loanTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_lending_redemption(self, redemption_time: int, lending_type: str, asset: str, amount: float,
                               auto_commit: bool = True):
        """
        add a lending redemption to the database

        :param redemption_time: millitstamp of the operation
        :type redemption_time: int
        :param lending_type: either 'DAILY', 'ACTIVITY' or 'CUSTOMIZED_FIXED'
        :type lending_type: str
        :param asset: asset lent
        :type asset: str
        :param amount: amount of asset redeemed
        :type amount: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (redemption_time, lending_type, asset, amount)
        self.add_row(tables.LENDING_REDEMPTION_TABLE, row, auto_commit=auto_commit)

    def get_lending_redemptions(self, lending_type: Optional[str] = None, asset: Optional[str] = None,
                                start_time: Optional[int] = None, end_time: Optional[int] = None):
        """
        return lending redemptions stored in the database. Asset type and time filters can be used

        :param lending_type: fetch only redemptions from this lending type
        :type lending_type: Optional[str]
        :param asset: fetch only redemptions from this asset
        :type asset: Optional[str]
        :param start_time: fetch only redemptions after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only redemptions before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                1612841562000,      # time
                'DAILY',            # lending type
                'LTC',              # asset
                1.89151684),        # amount
            ]
        """
        conditions_list = []
        table = tables.LENDING_REDEMPTION_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.redemptionTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.redemptionTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_lending_redemption_time(self, lending_type: Optional[str] = None) -> int:
        """
        return the latest time when an lending redemption was made.
        If None, return the millistamp corresponding to 2017/01/01

        :param lending_type: type of lending
        :type lending_type: str
        :return: millistamp
        :rtype: int
        """
        conditions_list = []
        table = tables.LENDING_REDEMPTION_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        selection = f"MAX({table.redemptionTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_lending_purchase(self, purchase_id: int, purchase_time: int, lending_type: str, asset: str, amount: float,
                             auto_commit: bool = True):
        """
        add a lending purchase to the database

        :param purchase_id: id of the purchase
        :type purchase_id: int
        :param purchase_time: millitstamp of the operation
        :type purchase_time: int
        :param lending_type: either 'DAILY', 'ACTIVITY' or 'CUSTOMIZED_FIXED'
        :type lending_type: str
        :param asset: asset lent
        :type asset: str
        :param amount: amount of asset lent
        :type amount: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (purchase_id, purchase_time, lending_type, asset, amount)
        self.add_row(tables.LENDING_PURCHASE_TABLE, row, auto_commit=auto_commit)

    def get_lending_purchases(self, lending_type: Optional[str] = None, asset: Optional[str] = None,
                              start_time: Optional[int] = None, end_time: Optional[int] = None):
        """
        return lending purchases stored in the database. Asset type and time filters can be used

        :param lending_type: fetch only purchases from this lending type
        :type lending_type: Optional[str]
        :param asset: fetch only purchases from this asset
        :type asset: Optional[str]
        :param start_time: fetch only purchases after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only purchases before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (58516828,          # purchase id
                1612841562000,      # time
                'DAILY',            # lending type
                'LTC',              # asset
                1.89151684),        # amount
            ]
        """
        conditions_list = []
        table = tables.LENDING_PURCHASE_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.purchaseTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.purchaseTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_lending_purchase_time(self, lending_type: Optional[str] = None) -> int:
        """
        return the latest time when an lending purchase was made.
        If None, return the millistamp corresponding to 2017/01/01

        :param lending_type: type of lending
        :type lending_type: str
        :return: millistamp
        :rtype: int
        """
        conditions_list = []
        table = tables.LENDING_PURCHASE_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        selection = f"MAX({table.purchaseTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_lending_interest(self, time: int, lending_type: str, asset: str, amount: float,
                             auto_commit: bool = True):
        """
        add an lending interest to the database

        :param time: millitstamp of the operation
        :type time: int
        :param lending_type: either 'DAILY', 'ACTIVITY' or 'CUSTOMIZED_FIXED'
        :type lending_type: str
        :param asset: asset that was received
        :type asset: str
        :param amount: amount of asset received
        :type amount: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (time, lending_type, asset, amount)
        self.add_row(tables.LENDING_INTEREST_TABLE, row, auto_commit=auto_commit)

    def get_lending_interests(self, lending_type: Optional[str] = None, asset: Optional[str] = None,
                              start_time: Optional[int] = None, end_time: Optional[int] = None):
        """
        return lending interests stored in the database. Asset type and time filters can be used

        :param lending_type: fetch only interests from this lending type
        :type lending_type: Optional[str]
        :param asset: fetch only interests from this asset
        :type asset: Optional[str]
        :param start_time: fetch only interests after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only interests before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (1619846515000,     # time
                'DAILY',            # lending type
                'DOT',              # asset
                0.00490156)         # amount
            ]

        """
        conditions_list = []
        table = tables.LENDING_INTEREST_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.interestTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.interestTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_lending_interest_time(self, lending_type: Optional[str] = None) -> int:
        """
        return the latest time when an interest was received.
        If None, return the millistamp corresponding to 2017/01/01

        :param lending_type: type of lending
        :type lending_type: str
        :return: millistamp
        :rtype: int
        """
        conditions_list = []
        table = tables.LENDING_INTEREST_TABLE
        if lending_type is not None:
            conditions_list.append((table.lendingType,
                                    SQLConditionEnum.equal,
                                    lending_type))
        selection = f"MAX({table.interestTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection,
                                          conditions_list=conditions_list)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_dust(self, tran_id: str, time: int, asset: str, asset_amount: float, bnb_amount: float, bnb_fee: float,
                 auto_commit: bool = True):
        """
        add dust operation to the database

        :param tran_id: id of the transaction (non unique)
        :type tran_id: str
        :param time: millitstamp of the operation
        :type time: int
        :param asset: asset that got converted to BNB
        :type asset: str
        :param asset_amount: amount of asset that got converted
        :type asset_amount: float
        :param bnb_amount: amount received from the conversion
        :type bnb_amount: float
        :param bnb_fee: fee amount in BNB
        :type bnb_fee: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """

        row = (tran_id, time, asset, asset_amount, bnb_amount, bnb_fee)
        self.add_row(tables.SPOT_DUST_TABLE, row, auto_commit=auto_commit)

    def get_spot_dusts(self, asset: Optional[str] = None, start_time: Optional[int] = None,
                       end_time: Optional[int] = None):
        """
        return dusts stored in the database. Asset type and time filters can be used

        :param asset: fetch only dusts from this asset
        :type asset: Optional[str]
        :param start_time: fetch only dusts after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only dusts before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (82156485284,       # transaction id
                1605489113400,      # time
                'TRX',              # asset
                102.78415879,       # asset amount
                0.09084498,         # bnb amount
                0.00171514),        # bnb fee
            ]
        """
        conditions_list = []
        table = tables.SPOT_DUST_TABLE
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.dustTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.dustTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def add_dividend(self, div_id: int, div_time: int, asset: str, amount: float, auto_commit: bool = True):
        """
        add a dividend to the database

        :param div_id: dividend id
        :type div_id: int
        :param div_time: millistamp of dividend reception
        :type div_time: int
        :param asset: name of the dividend unit
        :type asset: str
        :param amount: amount of asset distributed
        :type amount: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (div_id, div_time, asset, amount)
        self.add_row(tables.SPOT_DIVIDEND_TABLE, row, auto_commit=auto_commit)

    def get_spot_dividends(self, asset: Optional[str] = None, start_time: Optional[int] = None,
                           end_time: Optional[int] = None):
        """
        return dividends stored in the database. Asset type and time filters can be used

        :param asset: fetch only dividends of this asset
        :type asset: Optional[str]
        :param start_time: fetch only dividends after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only dividends before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (8945138941,         # dividend id
                1594513589000,       # time
                'TRX',               # asset
                0.18745654),         # amount
            ]

        """
        conditions_list = []
        table = tables.SPOT_DIVIDEND_TABLE
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.divTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.divTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_spot_dividend_time(self) -> int:
        """
        fetch the latest time a dividend has been distributed on the spot account. If None is found,
        return the millistamp corresponding to 2017/1/1

        :return:
        """
        table = tables.SPOT_DIVIDEND_TABLE
        selection = f"MAX({table.divTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection)

        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_withdraw(self, withdraw_id: str, tx_id: str, apply_time: int, asset: str, amount: float, fee: float,
                     auto_commit: bool = True):
        """
        add a withdraw to the database

        :param withdraw_id: binance if of the withdraw
        :type withdraw_id: str
        :param tx_id: transaction id
        :type tx_id: str
        :param apply_time: millistamp when the withdraw was requested
        :type apply_time: int
        :param asset: name of the token
        :type asset: str
        :param amount: amount of token withdrawn
        :type amount: float
        :param fee: amount of the asset paid for the withdraw
        :type fee: float
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (withdraw_id, tx_id, apply_time, asset, amount, fee)
        self.add_row(tables.SPOT_WITHDRAW_TABLE, row, auto_commit=auto_commit)

    def get_spot_withdraws(self, asset: Optional[str] = None, start_time: Optional[int] = None,
                           end_time: Optional[int] = None):
        """
        return withdraws stored in the database. Asset type and time filters can be used

        :param asset: fetch only withdraws of this asset
        :type asset: Optional[str]
        :param start_time: fetch only withdraws after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only withdraws before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                ('84984dcqq5z11gyjfa',  # withdraw id
                'aazd8949vredqs56dz',   # transaction id
                1599138389000,          # withdraw time
                'XTZ',                  # asset
                57.0194,                # amount
                0.5),                   # fee
            ]

        """
        conditions_list = []
        table = tables.SPOT_WITHDRAW_TABLE
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.applyTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.applyTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_spot_withdraw_time(self) -> int:
        """
        fetch the latest time a withdraw has been made on the spot account. If None is found, return the millistamp
        corresponding to 2017/1/1

        :return:
        """
        table = tables.SPOT_WITHDRAW_TABLE
        selection = f"MAX({table.applyTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection)
        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_deposit(self, tx_id: str, insert_time: int, amount: float, asset: str, auto_commit=True):
        """
        add a deposit to the database

        :param tx_id: transaction id
        :type tx_id: str
        :param insert_time: millistamp when the deposit arrived on binance
        :type insert_time: int
        :param amount: amount of token deposited
        :type amount: float
        :param asset: name of the token
        :type asset: str
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (tx_id, insert_time, asset, amount)
        self.add_row(tables.SPOT_DEPOSIT_TABLE, row, auto_commit)

    def get_spot_deposits(self, asset: Optional[str] = None, start_time: Optional[int] = None,
                          end_time: Optional[int] = None):
        """
        return deposits stored in the database. Asset type and time filters can be used

        :param asset: fetch only deposits of this asset
        :type asset: Optional[str]
        :param start_time: fetch only deposits after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only deposits before this millistamp
        :type end_time: Optional[int]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                ('azdf5e6a1d5z',    # transaction id
                1589479004000,      # deposit time
                'LTC',              # asset
                14.25),             # amount
            ]

        """
        conditions_list = []
        table = tables.SPOT_DEPOSIT_TABLE
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if start_time is not None:
            conditions_list.append((table.insertTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.insertTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        return self.get_conditions_rows(table, conditions_list=conditions_list)

    def get_last_spot_deposit_time(self) -> int:
        """
        fetch the latest time a deposit has been made on the spot account. If None is found, return the millistamp
        corresponding to 2017/1/1

        :return: last deposit millistamp
        :rtype: int
        """
        table = tables.SPOT_DEPOSIT_TABLE
        selection = f"MAX({table.insertTime})"
        result = self.get_conditions_rows(table,
                                          selection=selection)

        default = datetime_to_millistamp(datetime.datetime(2017, 1, 1, tzinfo=datetime.timezone.utc))
        try:
            result = result[0][0]
        except IndexError:
            return default
        if result is None:
            return default
        return result

    def add_trade(self, trade_type: str, trade_id: int, trade_time: int, asset: str, ref_asset: str, qty: float,
                  price: float, fee: float, fee_asset: str, is_buyer: bool, auto_commit=True):
        """
        add a trade to the database

        :param trade_type: type trade executed
        :type trade_type: string, must be one of {'spot', 'cross_margin'}
        :param trade_id: id of the trade (binance id, unique per trading pair)
        :type trade_id: int
        :param trade_time: millistamp of the trade
        :type trade_time: int
        :param asset: name of the asset in the trading pair (ex 'BTC' for 'BTCUSDT')
        :type asset: string
        :param ref_asset: name of the reference asset in the trading pair (ex 'USDT' for 'BTCUSDT')
        :type ref_asset: string
        :param qty: quantity of asset exchanged
        :type qty: float
        :param price: price of the asset regarding the ref_asset
        :type price: float
        :param fee: amount kept by the exchange
        :type fee: float
        :param fee_asset: token unit for the fee
        :type fee_asset: str
        :param is_buyer: if the trade is a buy or a sell
        :type is_buyer: bool
        :param auto_commit: if the database should commit the change made, default True
        :type auto_commit: bool
        :return: None
        :rtype: None
        """
        row = (trade_id, trade_time, asset, ref_asset, qty, price, fee, fee_asset, int(is_buyer))
        if trade_type == 'spot':
            table = tables.SPOT_TRADE_TABLE
        elif trade_type == 'cross_margin':
            table = tables.CROSS_MARGIN_TRADE_TABLE
        else:
            raise ValueError(f"trade type should be one of ('spot', 'cross_margin') but {trade_type} was received")
        self.add_row(table, row, auto_commit)

    def get_trades(self, trade_type: str, start_time: Optional[int] = None, end_time: Optional[int] = None,
                   asset: Optional[str] = None, ref_asset: Optional[str] = None):
        """
        return trades stored in the database. asset type, ref_asset type and time filters can be used

        :param trade_type: type trade executed
        :type trade_type: string, must be one of ('spot', 'cross_margin')
        :param start_time: fetch only trades after this millistamp
        :type start_time: Optional[int]
        :param end_time: fetch only trades before this millistamp
        :type end_time: Optional[int]
        :param asset: fetch only trades with this asset
        :type asset: Optional[str]
        :param ref_asset:  fetch only trades with this ref_asset
        :type ref_asset: Optional[str]
        :return: The raw rows selected as saved in the database
        :rtype: List[Tuple]

        .. code-block:: python

            [
                (384518832,         # trade_id
                1582892988052,      # trade time
                'BTC',              # asset
                'USDT',             # ref asset
                0.0015,             # asset quantity
                9011.2,             # asset price to ref asset
                0.01425,            # fee
                'USDT',             # fee asset
                0),                 # is_buyer
            ]

        """
        if trade_type == 'spot':
            table = tables.SPOT_TRADE_TABLE
        elif trade_type == 'cross_margin':
            table = tables.CROSS_MARGIN_TRADE_TABLE
        else:
            raise ValueError(f"trade type should be one of ('spot', 'cross_margin') but {trade_type} was received")
        conditions_list = []
        if start_time is not None:
            conditions_list.append((table.tdTime,
                                    SQLConditionEnum.greater_equal,
                                    start_time))
        if end_time is not None:
            conditions_list.append((table.tdTime,
                                    SQLConditionEnum.lower,
                                    end_time))
        if asset is not None:
            conditions_list.append((table.asset,
                                    SQLConditionEnum.equal,
                                    asset))
        if ref_asset is not None:
            conditions_list.append((table.refAsset,
                                    SQLConditionEnum.equal,
                                    ref_asset))
        return self.get_conditions_rows(table, conditions_list=conditions_list, order_list=[table.tdTime])

    def get_max_trade_id(self, asset: str, ref_asset: str, trade_type: str) -> int:
        """
        return the latest trade id for a trading pair. If none is found, return -1

        :param asset: name of the asset in the trading pair (ex 'BTC' for 'BTCUSDT')
        :type asset: string
        :param ref_asset: name of the reference asset in the trading pair (ex 'USDT' for 'BTCUSDT')
        :type ref_asset: string
        :param trade_type: type trade executed
        :type trade_type: string, must be one of {'spot', 'cross_margin'}
        :return: latest trade id
        :rtype: int
        """
        if trade_type == 'spot':
            table = tables.SPOT_TRADE_TABLE
        elif trade_type == 'cross_margin':
            table = tables.CROSS_MARGIN_TRADE_TABLE
        else:
            raise ValueError(f"trade type should be one of {'spot', 'cross_margin'} but {trade_type} was received")

        selection = f"MAX({table.tradeId})"
        conditions_list = [
            (table.asset,
             SQLConditionEnum.equal,
             asset),
            (table.refAsset,
             SQLConditionEnum.equal,
             ref_asset)
        ]
        result = self.get_conditions_rows(table, selection=selection, conditions_list=conditions_list)
        try:
            result = result[0][0]
        except IndexError:
            return -1
        if result is None:
            return -1
        return result



import datetime


def millistamp_to_round_min(millitstamp: int):
    return millitstamp - millitstamp % 60000


def millistamp_to_upper_min(millitstamp: int):
    over_min = millitstamp % 60000
    if over_min:
        return millitstamp - over_min + 60000
    return millitstamp


def datetime_to_round_min(date_time: datetime.datetime):
    return date_time - datetime.timedelta(seconds=date_time.second, microseconds=date_time.microsecond)


def millistamp_to_datetime(millitstamp: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(millitstamp / 1000, tz=datetime.timezone.utc)


def datetime_to_millistamp(date_time: datetime.datetime) -> int:
    return int(1000 * date_time.timestamp())



from dataclasses import dataclass
from typing import List, Optional


class Table:
    """
    @DynamicAttrs
    """

    def __init__(self, name: str, columns_names: List[str], columns_sql_types: List[str],
                 primary_key: Optional[str] = None, primary_key_sql_type: Optional[str] = None):
        self.name = name
        self.columns_names = columns_names
        self.columns_sql_types = columns_sql_types
        self.primary_key = primary_key
        self.primary_key_sql_type = primary_key_sql_type

        for column_name in self.columns_names:
            try:
                value = getattr(self, column_name)
                raise ValueError(f"the name {column_name} conflicts with an existing attribute of value {value}")
            except AttributeError:
                setattr(self, column_name, column_name)

        if self.primary_key is not None:
            setattr(self, self.primary_key, self.primary_key)


SPOT_TRADE_TABLE = Table(
    'spot_trade',
    [
        'tradeId',
        'tdTime',
        'asset',
        'refAsset',
        'qty',
        'price',
        'fee',
        'feeAsset',
        'isBuyer'

    ],
    [
        'INTEGER',
        'INTEGER',
        'TEXT',
        'TEXT',
        'REAL',
        'REAL',
        'REAL',
        'TEXT',
        'INTEGER'
    ]
)

SPOT_DEPOSIT_TABLE = Table(
    'spot_deposit',
    [
        'insertTime',
        'asset',
        'amount',
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL'
    ],
    primary_key='txId',
    primary_key_sql_type='TEXT'
)


SPOT_WITHDRAW_TABLE = Table(
    'spot_withdraw',
    [
        'txId',
        'applyTime',
        'asset',
        'amount',
        'fee'
    ],
    [
        'TEXT',
        'INTEGER',
        'TEXT',
        'REAL',
        'REAL'
    ],
    primary_key='withdrawId',
    primary_key_sql_type='TEXT'
)

SPOT_DIVIDEND_TABLE = Table(
    'spot_dividend_table',
    [
        'divTime',
        'asset',
        'amount'
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL'
    ],
    primary_key='divId',
    primary_key_sql_type='INTEGER'
)

SPOT_DUST_TABLE = Table(
    'spot_dust_table',
    [
        'tranId',
        'dustTime',
        'asset',
        'assetAmount',
        'bnbAmount',
        'bnbFee',
    ],
    [
        'INTEGER',
        'INTEGER',
        'TEXT',
        'REAL',
        'REAL',
        'REAL'
    ]
)

LENDING_INTEREST_TABLE = Table(
    'lending_interest_table',
    [
        'interestTime',
        'lendingType',
        'asset',
        'amount',
    ],
    [
        'INTEGER',
        'TEXT',
        'TEXT',
        'REAL',
    ]
)

LENDING_PURCHASE_TABLE = Table(
    'lending_purchase_history',
    [
        'purchaseTime',
        'lendingType',
        'asset',
        'amount'
    ],
    [
        'INTEGER',
        'TEXT',
        'TEXT',
        'INTEGER'
    ],
    primary_key='purchaseId',
    primary_key_sql_type='INTEGER'
)

LENDING_REDEMPTION_TABLE = Table(
    'lending_redemption_history',
    [
        'redemptionTime',
        'lendingType',
        'asset',
        'amount'
    ],
    [
        'INTEGER',
        'TEXT',
        'TEXT',
        'INTEGER'
    ]
)

CROSS_MARGIN_TRADE_TABLE = Table(
    'cross_margin_trade',
    [
        'tradeId',
        'tdTime',
        'asset',
        'refAsset',
        'qty',
        'price',
        'fee',
        'feeAsset',
        'isBuyer'

    ],
    [
        'INTEGER',
        'INTEGER',
        'TEXT',
        'TEXT',
        'REAL',
        'REAL',
        'REAL',
        'TEXT',
        'INTEGER'
    ]
)

CROSS_MARGIN_LOAN_TABLE = Table(
    "cross_margin_loan_table",
    [
        'loanTime',
        'asset',
        'principal',
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL'
    ],
    primary_key='txId',
    primary_key_sql_type='INTEGER'

)

CROSS_MARGIN_REPAY_TABLE = Table(
    "cross_margin_repay_table",
    [
        'repayTime',
        'asset',
        'principal',
        'interest',
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL',
        'REAL'
    ],
    primary_key='txId',
    primary_key_sql_type='INTEGER'

)

CROSS_MARGIN_INTEREST_TABLE = Table(
    "cross_margin_interest_table",
    [
        'interestTime',
        'asset',
        'interest',
        'interestType'
    ],
    [
        'INTEGER',
        'TEXT',
        'REAL',
        'TEXT'
    ]
)

UNIVERSAL_TRANSFER_TABLE = Table(
    "universal_transfer_table",
    [
        'trfType',
        'trfTime',
        'asset',
        'amount'
    ],
    [
        'TEXT',
        'INTEGER',
        'TEXT',
        'REAL'
    ],
    primary_key='tranId',
    primary_key_sql_type='INTEGER'
)



import argparse
import traceback
from types import ModuleType
from typing import List, Optional, Dict

import tests.print_utils as pu
import tests


def get_test_modules(module_names: Optional[List[str]] = None) -> List[ModuleType]:
    """
    from a list of module names, will return the list of module objects belonging to the tests package
    if the list is empty, will return all the modules with 'test_' in the name
    :param module_names: list of modules from the tests package to fetch
    :return:
    """
    if module_names is None:
        module_names = [m for m in dir(tests) if 'test_' in m]
    test_modules = []
    for module_name in module_names:
        try:
            test_modules.append(getattr(tests, module_name))
        except AttributeError:
            message = f"The test module {module_name} was skipped because the file was not found in the tests package"
            print(pu.s_to_warning(message))
    return test_modules


def run_test_module(module: ModuleType, common_kwargs=None, verbose=1) -> Dict:
    """
    will run all the functions from a module whose name has 'test_' in it and return an execution report
    :param module: module that host the functions
    :param common_kwargs: kwargs to provide to all functions
    :param verbose: the higher the more information are printed
    :return: dictionary with execution report ex: {'success': 10, 'failure': 1, 'total': 11}
    """
    if common_kwargs is None:
        common_kwargs = {}
    n_success = 0
    n_failure = 0
    tests_names = [f_name for f_name in dir(module) if 'test_' in f_name and callable(getattr(module, f_name))]
    for test_name in tests_names:
        if verbose:
            print(pu.get_blue_title(module.__name__, test_name))
        try:
            getattr(module, test_name)(**common_kwargs)
            n_success += 1
        except Exception:  # general catch to display the error reports without stopping the tests
            n_failure += 1
            if verbose:
                print(pu.s_to_fail(traceback.format_exc()))
            pass
        if verbose:
            print(pu.get_blue_sep())
    return {'success': n_success, 'failure': n_failure, 'total': len(tests_names)}


def print_report(report: Dict):
    """
    print nicely a report of test execution
    :param report:
    """
    print("\n\nSummary:\n")
    total_success = 0
    total_tests = 0
    for module_name, dico in report.items():
        if dico['total']:
            total_success += dico['success']
            total_tests += dico['total']
            rate = dico['success'] / dico['total']
            if rate == 1:
                s_result = pu.s_to_pass(f"{dico['success'] / dico['total']:.0%}")
            else:
                s_result = pu.s_to_fail(f"{dico['success'] / dico['total']:.0%}")
            print(f"\t- {module_name}: {dico['success']}/{dico['total']} ({s_result}) tests were passed")
        else:
            print(f"\t- {module_name}: no tests were executed")


def run(tests_names=None, common_kwargs=None):
    """
    main function, launch the tests function from a list of test modules names
    :param tests_names: testers to execute
    :param common_kwargs: kwargs for all modules' functions
    :return:
    """
    if common_kwargs is None:
        common_kwargs = {}
    test_modules = get_test_modules(tests_names)

    report = {}
    for test_module in test_modules:
        report[test_module.__name__] = run_test_module(test_module, common_kwargs)

    if len(report):
        print_report(report)
    else:
        print("No report to print")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--test_modules', metavar="<test_modules>", required=False,
                        help='names of the test modules to be executed', type=str, nargs='+')
    parser.add_argument('-v', '--verbose', metavar="<verbose>", required=False,
                        help='verbose: the higher the more information are printed', type=int)

    args = parser.parse_args()

    testers_args = {'verbose': 0}
    if args.verbose:
        testers_args['verbose'] = args.verbose

    run(args.test_modules, testers_args)



import os
from setuptools import setup

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

about = {}
with open(os.path.join(this_directory, 'BinanceWatch/__init__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

setup(
    name='BinanceWatch',
    version=about['__version__'],
    packages=['BinanceWatch',
              'tests',
              'BinanceWatch.storage',
              'BinanceWatch.utils'],
    url='https://github.com/EtWnn/BinanceWatch',
    author='EtWnn',
    author_email='',
    license='MIT',
    description='Local tracker of a binance account',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    install_requires=['numpy', 'tqdm', 'dateparser', 'requests', 'python-binance>=0.7.9', 'appdirs'],
    keywords='binance exchange wallet save tracking history bitcoin ethereum btc eth',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
